/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*-
   this file is part of rcssserver3D
   Thu Mar 27 2006
   Copyright (C) 2006 RoboCup Simulation League Maintenance Committee

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef VECPOSITION_H
#define VECPOSITION_H

#include "math.h"       // needed for M_PI constant
#include <string.h>       // needed for string
#include <iostream>

#include "headers.h"

using namespace std;

/*! CoordSystem is an enumeration of the different specified
    coordinate systems.  The two possibilities are CARTESIAN or
    POLAR. These values are for instance used in the initializing a
    VecPosition. The CoordSystem indicates whether the supplied
    arguments represent the position in cartesian or in polar
    coordinates. */

enum CoordSystemT {
  CARTESIAN = 0,
  POLAR = 1
};

// auxiliary goniometric functions which enable you to
// specify angles in degrees rather than in radians
double Rad2Deg ( double x             );
double Deg2Rad ( double x             );
double cosDeg  ( double x             );
double sinDeg  ( double x             );
double tanDeg  ( double x             );
double atanDeg ( double x             );
double atan2Deg( double x,  double y  );
double acosDeg ( double x             );
double asinDeg ( double x             );


class VecPosition{
  
  // private member data
 private:
  
  double m_x;   /*!< x-coordinate of this position */
  double m_y;   /*!< y-coordinate of this position */
  double m_z;   /*!< z-coordinate of this position */
  
  // public methods
 public:
  // constructor for VecPosition class
  VecPosition(double vx = 0, double vy = 0, double vz = 0, CoordSystemT cs = CARTESIAN);
  
  // overloaded arithmetic operators
  VecPosition        operator -             (                                );
  VecPosition        operator +             ( const double      &d           );
  VecPosition        operator +             ( const VecPosition &p           );
  VecPosition        operator -             ( const double      &d           );
  VecPosition        operator -             ( const VecPosition &p           );
  VecPosition        operator *             ( const double      &d           );
  VecPosition        operator *             ( const VecPosition &p           );
  VecPosition        operator /             ( const double      &d           );
  VecPosition        operator /             ( const VecPosition &p           );
  void               operator =             ( const double      &d           );
  void               operator +=            ( const VecPosition &p           );
  void               operator +=            ( const double      &d           );
  void               operator -=            ( const VecPosition &p           );
  void               operator -=            ( const double      &d           );
  void               operator *=            ( const VecPosition &p           );
  void               operator *=            ( const double      &d           );
  void               operator /=            ( const VecPosition &p           );
  void               operator /=            ( const double      &d           );
  bool               operator !=            ( const VecPosition &p           );
  bool               operator !=            ( const double      &d           );
  bool               operator ==            ( const VecPosition &p           );
  bool               operator ==            ( const double      &d           );


  // methods for producing output
    friend ostream&    operator <<            ( ostream           &os,
                                                  VecPosition       p            );

  void               show                   ( CoordSystemT      cs =CARTESIAN);

  // set- and get methods for private member variables
  bool               setX                   ( double            dX           );
  double             getX                   (                          ) const;
  bool               setY                   ( double            dY           );
  double             getY                   (                          ) const;
  bool               setZ                   ( double            dZ           );
  double             getZ                   (                          ) const;

  // set- and get methods for derived position information
  void               setVecPosition         ( double            dX = 0,
                                              double            dY = 0,
                                              double            dZ = 0,
                                              CoordSystemT      cs =CARTESIAN);
  double             getDistanceTo          ( const VecPosition p            );
  VecPosition        setMagnitude           ( double            d            );
  double             getMagnitude           (                          ) const;


  double             getTheta           (                          ) const;
  double             getPhi           (                          ) const;

  VecPosition        normalize              (                                );

    double             getAngleBetweenPoints  ( const VecPosition &p1,
				              const VecPosition &p2          );

    //static class methods
    static VecPosition getVecPositionFromPolar(double r, double theta, double phi);
    
    VecPosition getCartesianFromPolar();
    
    static double     normalizeAngle         ( double            angle        );

    VecPosition translate(VecPosition t);
    VecPosition rotateAboutX(double angle);
    VecPosition rotateAboutY(double angle);
    VecPosition rotateAboutZ(double angle);
    double dotProduct(VecPosition v);
    double getAngleWithVector(VecPosition v);
    VecPosition crossProduct(VecPosition v);


};


#endif // VECPOSITION_H

