#include "naobehavior.h"

/*
  This function returns true if we are taking some action 
  to avert or recover from fall; false otherwise.
 
  If fallState is 0, we are still safe, and a check is performed
  to detect fall along four directions: up, down, right, and left.
  If any of these indicates fall, the fallState becomes 1, and the robot
  stretches out its arms sideways. This action typically forces the robot 
  into a up or down fallen position, i.e., not sideways. fallState 1 leads 
  to fallState 2, when the fall direction is reevaluated. 

  From fallenUp or fallenDown, fallState 3 leads to a sequence of fallStates
  that is ultimately reset to 0.
 */

bool NaoBehavior::checkingFall(){

  VecPosition accel = bodyModel->getAccelRates();
  if(fallState == 0){
   // Check if falling. Only one flag is to be set if falling.
    fallenUp = (accel.getX() < -6.5);
    fallenDown = !fallenUp && (accel.getX() > 6.5);
    fallenRight = !fallenUp && !fallenDown && (accel.getY() < -6.5);
    fallenLeft = !fallenUp && !fallenDown && !fallenRight && (accel.getY() >  6.5);
    
    if(fallenLeft){
      //cout << "----------FALLEN LEFT-------------\n";
    }
    else if(fallenRight){
      //cout << "----------FALLEN RIGHT-------------\n";
    }
    else if(fallenUp){
      //cout << "----------FALLEN UP-------------\n";
    }
    else if(fallenDown){
      //cout << "----------FALLEN DOWN-------------\n";
    }
    
    if(fallenUp || fallenDown || fallenLeft || fallenRight){
      fallState = 1;
      worldModel->setFallen(true);
    }
    else{
      worldModel->setFallen(false);
      return false;
    }
  }

  
  // Don't get stuck in getup
  if (fallTimeWait >= 0 && (worldModel->getTime() - fallTimeWait > 3)) {
    // Both fall get ups have states 1,3,5,7, and 9 which change movement so 
    // choose one of these states
    fallState = (rand()%5)*2+1;
    fallTimeWait = -1;
    //cout << "Resetting fall state" << endl;
  }

  //cout << "Fall State: " << fallState << "\n";

  if(fallState == 1){

      bodyModel->setInitialArm(ARM_LEFT);
      bodyModel->setInitialArm(ARM_RIGHT);
      bodyModel->setInitialLeg(LEG_LEFT);
      bodyModel->setInitialLeg(LEG_RIGHT);
      
      bodyModel->setTargetAngle(EFF_LA2, 90);
      bodyModel->setTargetAngle(EFF_RA2, -90);

      bodyModel->setTargetAngle(EFF_LA4, 0);
      bodyModel->setTargetAngle(EFF_RA4, 0);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      if(bodyModel->targetsReached()){
	fallState = 2;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
  }
  else if(fallState == 2){

    if(worldModel->getTime() > (fallTimeStamp + 0.5)){
      fallTimeStamp = -1;
      fallState = 1;

      // Check once again. If still fallen right or left, act as though fallen up.
      fallenUp = (accel.getX() < -6.5);
      fallenDown = !fallenUp && (accel.getX() > 6.5);
      fallenRight = !fallenUp && !fallenDown && (accel.getY() < -6.5);
      fallenLeft = !fallenUp && !fallenDown && !fallenRight && (accel.getY() >  6.5);

      if(fallenUp || fallenRight || fallenLeft){
	fallenUp = true;
	fallenRight = false;
	fallenLeft = false;
	fallState = 3;
      }
      else if(fallenDown){
	fallState = 3;
      }
      else{//Magically, we are up!
	fallState = 0;
      }
    }
  }


  //Recovering
  if(fallenDown){

    if(fallState == 3){

      bodyModel->setTargetAngle(EFF_LA1, 30);
      bodyModel->setTargetAngle(EFF_RA1, 30);

      bodyModel->setTargetAngle(EFF_LA2, 0);
      bodyModel->setTargetAngle(EFF_RA2, 0);

      bodyModel->setTargetAngle(EFF_LL3, 110.0);
      bodyModel->setTargetAngle(EFF_RL3, 110.0);
      
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }

      if(bodyModel->targetsReached()){
	fallState = 4;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 4){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 5;
      }
    }
    else if(fallState == 5){

      bodyModel->setTargetAngle(EFF_LL1, -45);
      bodyModel->setTargetAngle(EFF_RL1, -45);

      bodyModel->setTargetAngle(EFF_LL5, 0);
      bodyModel->setTargetAngle(EFF_RL5, 0);
      
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 6;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 6){
      if(worldModel->getTime() > (fallTimeStamp + 0.25)){
	fallTimeStamp = -1;
	fallState = 7;
      }
    }
    else if(fallState == 7){

      bodyModel->setTargetAngle(EFF_LL1, -25);
      bodyModel->setTargetAngle(EFF_RL1, -25);

      bodyModel->setTargetAngle(EFF_LL3, 60);
      bodyModel->setTargetAngle(EFF_RL3, 60);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 8;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 8){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 9;
      }
    }
    else if(fallState == 9){

      bodyModel->setInitialArm(ARM_LEFT);
      bodyModel->setInitialArm(ARM_RIGHT);
      bodyModel->setInitialLeg(LEG_LEFT);
      bodyModel->setInitialLeg(LEG_RIGHT);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }

      if(bodyModel->targetsReached()){
	fallState = 10;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 10){
      if(worldModel->getTime() > (fallTimeStamp + 0.1)){
	fallTimeStamp = -1;
	fallState = 11;
	fallState = 0;
	fallenDown = false;
      }
    }
  }

  if(fallenUp){

    if(fallState == 3){

      bodyModel->setTargetAngle(EFF_LA1, -120.0);
      bodyModel->setTargetAngle(EFF_RA1, -120.0);

      bodyModel->setTargetAngle(EFF_LA2, 35.0);
      bodyModel->setTargetAngle(EFF_RA2, -35.0);

      bodyModel->setTargetAngle(EFF_LA4, 60.0);
      bodyModel->setTargetAngle(EFF_RA4, -60.0);

      bodyModel->setTargetAngle(EFF_LL3, 15.0);
      bodyModel->setTargetAngle(EFF_RL3, 15.0);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      if(bodyModel->targetsReached()){
	fallState = 4;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 4){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 5;
      }
    }
    else if(fallState == 5){

      bodyModel->setTargetAngle(EFF_LL3, 110.0);
      bodyModel->setTargetAngle(EFF_RL3, 110.0);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }

      if(bodyModel->targetsReached()){
	fallState = 6;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 6){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 7;
      }
    }
    else if(fallState == 7){

      bodyModel->setTargetAngle(EFF_LA1, 0);
      bodyModel->setTargetAngle(EFF_RA1, 0);

      bodyModel->setTargetAngle(EFF_LA2, 0);
      bodyModel->setTargetAngle(EFF_RA2, 0);

      bodyModel->setTargetAngle(EFF_LL1, -35.0);
      bodyModel->setTargetAngle(EFF_RL1, -35.0);
      
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 8;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }

    }
    else if(fallState == 8){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 9;
      }
    }
    else if(fallState == 9){

      bodyModel->setTargetAngle(EFF_LA1, 30);
      bodyModel->setTargetAngle(EFF_RA1, 30);

      bodyModel->setTargetAngle(EFF_LL1, -80.0);
      bodyModel->setTargetAngle(EFF_RL1, -80.0);

      bodyModel->setTargetAngle(EFF_LL4, -60.0);
      bodyModel->setTargetAngle(EFF_RL4, -60.0);

      bodyModel->setTargetAngle(EFF_LL5, -70.0);
      bodyModel->setTargetAngle(EFF_RL5, -70.0);

      bodyModel->setTargetAngle(EFF_LL6, -40.0);
      bodyModel->setTargetAngle(EFF_RL6, 40.0);
      
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 10;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 10){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 11;
      }
    }
    else if(fallState == 11){

      bodyModel->setTargetAngle(EFF_LA1, 30);
      bodyModel->setTargetAngle(EFF_RA1, 30);

      bodyModel->setTargetAngle(EFF_LL1, -60.0);
      bodyModel->setTargetAngle(EFF_RL1, -60.0);

      bodyModel->setTargetAngle(EFF_LL5, -80);
      bodyModel->setTargetAngle(EFF_RL5, -80);
     
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }

      if(bodyModel->targetsReached()){
	fallState = 12;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 12){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 13;
      }
    }
    else if(fallState == 13){

      bodyModel->setTargetAngle(EFF_LA1, -90);
      bodyModel->setTargetAngle(EFF_RA1, -90);

      bodyModel->setTargetAngle(EFF_LL1, -30.0);
      bodyModel->setTargetAngle(EFF_RL1, -30.0);

      bodyModel->setTargetAngle(EFF_LL3, 30.0);
      bodyModel->setTargetAngle(EFF_RL3, 30.0);

      bodyModel->setTargetAngle(EFF_LL4, -10);
      bodyModel->setTargetAngle(EFF_RL4, -10);

      bodyModel->setTargetAngle(EFF_LL5, -10);
      bodyModel->setTargetAngle(EFF_RL5, -10);

      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 14;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 14){
      if(worldModel->getTime() > (fallTimeStamp + 0)){
	fallTimeStamp = -1;
	fallState = 15;
      }
    }
    else if(fallState == 15){

      bodyModel->setInitialArm(ARM_LEFT);
      bodyModel->setInitialArm(ARM_RIGHT);
      bodyModel->setInitialLeg(LEG_LEFT);
      bodyModel->setInitialLeg(LEG_RIGHT);
      
      if (fallTimeWait < 0) {
	fallTimeWait = worldModel->getTime();
      }
      
      if(bodyModel->targetsReached()){
	fallState = 16;
	fallTimeStamp = worldModel->getTime();
	fallTimeWait = -1;
      }
    }
    else if(fallState == 16){
      if(worldModel->getTime() > (fallTimeStamp + 0.1)){
	fallTimeStamp = -1;
	fallState = 0;
	fallenUp = false;
      }
    }
  }

  return true;

}

