#include "skill.h"


/////////////////////////////////////////////////////
// Macros
/////////////////////////////////////////////////////

// IncTar
// -------
IncTar::IncTar(const vector<int> &effectorIDs_, const vector<double> &increments_) :
  effectorIDs(effectorIDs_), increments(increments_) {
  // assuming two vectors are the same length
}


void IncTar::execute(BodyModel *bodyModel){
  // assuming two vectors are the same length
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    bodyModel->increaseTargetAngle( effectorIDs[i], increments[i] );
  }
}

shared_ptr<Macro> 
IncTar::getReflection(BodyModel *bodyModel){

  vector<int> newEffIDs;
  vector<double> newIncs;

  for( int i = 0; i < effectorIDs.size(); ++i ) {
    int newEffID;
    double newIncrement;
    bodyModel->getReflection( effectorIDs[i], 
                              increments[i], 
                              newEffID, 
                              newIncrement );
    newEffIDs.push_back( newEffID );
    newIncs.push_back( newIncrement );
  }

  return shared_ptr<Macro>( new IncTar(newEffIDs, newIncs) );
}

void IncTar::display() {
  cerr << "Displaying macro: IncTar" << endl;
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    cerr << "effID[" << i << "]=" << effectorIDs[i] 
         << " increment[" << i << "]=" << increments[i] << endl;   
  }
}

// SetTar
// -------
SetTar::SetTar(const vector<int> &effectorIDs_, const vector<double> &targetAngles_) :
  effectorIDs(effectorIDs_), targetAngles(targetAngles_) {
  // assuming two vectors are the same length
}


void SetTar::execute(BodyModel *bodyModel){
  // assuming two vectors are the same length
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    bodyModel->setTargetAngle( effectorIDs[i], targetAngles[i] );
  }
}

shared_ptr<Macro> 
SetTar::getReflection(BodyModel *bodyModel){

  vector<int> newEffIDs;
  vector<double> newAngs;

  for( int i = 0; i < effectorIDs.size(); ++i ) {
    int newEffID;
    double newAngle;
    bodyModel->getReflection( effectorIDs[i], 
                              targetAngles[i], 
                              newEffID, 
                              newAngle );
    newEffIDs.push_back( newEffID );
    newAngs.push_back( newAngle );
  }

  return shared_ptr<Macro>( new SetTar(newEffIDs, newAngs) );
}

void SetTar::display() {
  cerr << "Displaying macro: SetTar" << endl;
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    cerr << "effID[" << i << "]=" << effectorIDs[i] 
         << " Angle[" << i << "]=" << targetAngles[i] << endl;   
  }
}

// Reset
// -----
Reset::Reset(const std::vector<int>& components_) :
  components( components_ ) {
}

void Reset::
execute(BodyModel *bodyModel) {
  vector<int>::iterator begin( components.begin() ), end( components.end() );
  for( ; begin != end; ++begin ) {
    int currentJoint = *begin;
    if( currentJoint == HEAD ) {
      bodyModel->setInitialHead();
    } else if( currentJoint == ARM_LEFT ) {
      bodyModel->setInitialArm(ARM_LEFT);
    } else if( currentJoint == ARM_RIGHT ) {
      bodyModel->setInitialArm(ARM_RIGHT);
    } else if( currentJoint == LEG_LEFT ) {
      bodyModel->setInitialLeg(LEG_LEFT);
    } else if( currentJoint == LEG_RIGHT ) {
      bodyModel->setInitialLeg(LEG_RIGHT);
    }
  }
}

shared_ptr<Macro> 
Reset::getReflection(BodyModel *bodyModel) {
  
  vector<int> newComponents = components;

  for(unsigned i = 0; i < components.size(); ++i ) {

    if( components[i] == ARM_LEFT ) {
      newComponents[i] = ARM_RIGHT;    
    } else if ( components[i] == ARM_RIGHT ) {
      newComponents[i] = ARM_LEFT;
    } else if ( components[i] == LEG_LEFT ) {
      newComponents[i] = LEG_RIGHT;
    } else if ( components[i] == LEG_RIGHT ) {
      newComponents[i] = LEG_LEFT;
    }
  }
  return shared_ptr<Macro>( new Reset( newComponents ) );
}

void Reset::
display() {
  cerr << "Displaying macro: Reset" << endl;
  vector<int>::iterator it( components.begin() ), end( components.end() );
  for( ; it != end; ++it ) {
    cerr << *it << ' ';
  }
  cerr << endl;
}


// SetScale
// --------

SetScale::SetScale(const vector<int> &effectorIDs_, const vector<double> &targetScales_) :
  effectorIDs(effectorIDs_), targetScales(targetScales_) {
  // assuming two vectors are the same length
}


void SetScale::execute(BodyModel *bodyModel){
  // assuming two vectors are the same length
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    bodyModel->setScale( effectorIDs[i], targetScales[i] );
  }
}

shared_ptr<Macro> 
SetScale::getReflection(BodyModel *bodyModel){

  vector<int> newEffIDs;

  for( int i = 0; i < effectorIDs.size(); ++i ) {
    int newEffID;
    double dummyAngle1, dummyAngle2;
    bodyModel->getReflection( effectorIDs[i], 
                              dummyAngle1, 
                              newEffID, 
                              dummyAngle2 );
    newEffIDs.push_back( newEffID );
  }

  return shared_ptr<Macro>( new SetScale(newEffIDs, targetScales) );
}

void SetScale::display() {
  cerr << "Displaying macro: SetScale" << endl;
  for( int i = 0; i < effectorIDs.size(); ++i ) {
    cerr << "effID[" << i << "]=" << effectorIDs[i] 
         << " Scale[" << i << "]=" << targetScales[i] << endl;   
  }
}



/////////////////////////////////////////////////////////
// KeyFrame
/////////////////////////////////////////////////////////
KeyFrame::KeyFrame() {
  // sets the default values
  toWaitTime = false;
  waitTime = 0;
  toWaitTargets = false;
  maxWaitTime = INF;
}

bool KeyFrame::done(BodyModel *bodyModel, const double &setTime, const double &currentTime){

  // if invalid time - not done
  if( setTime < 0 ) {
    return false;
  }

  // Exceptional case. Break out of keyframe always.
  if((currentTime - setTime) >= maxWaitTime){
    return true;
  }

  // If we have to wait time, have we waited long enough?
  double epsilon = 1e-4;
  if(toWaitTime && (currentTime - setTime) < waitTime - epsilon ){
    return false;
  }

  // If we have to wait for targets to be reached, have they been reached?
  if(toWaitTargets && !(bodyModel->targetsReached())){
    return false;
  }

  // We have finished time (if at all) and reached targets (if at all).
  return true;
}

void KeyFrame::execute(BodyModel *bodyModel){

  for(unsigned m = 0; m < macros.size(); m++){
    macros[m]->execute(bodyModel);
  }
}

void KeyFrame::setToWaitTime( bool value ) {
  toWaitTime = value;
}

void KeyFrame::setWaitTime( double value ) {
  waitTime = value;
}

void KeyFrame::setToWaitTargets( bool value ) {
  toWaitTargets = value;
}

void KeyFrame::setMaxWaitTime( double value ) {
  maxWaitTime = value;
}

void KeyFrame::appendMacro(shared_ptr<Macro> macro) {
  macros.push_back( macro );
}

shared_ptr<KeyFrame> KeyFrame::getReflection(BodyModel *bodyModel){

  shared_ptr<KeyFrame> reflection(new KeyFrame());

  reflection->setToWaitTime(this->toWaitTime);
  reflection->setWaitTime(this->waitTime);
  reflection->setToWaitTargets(this->toWaitTargets);
  reflection->setMaxWaitTime(this->maxWaitTime);

  (reflection->macros).clear();
  for(int m = 0; m < this->macros.size(); m++){
    reflection->appendMacro(this->macros[m]->getReflection(bodyModel));
  }

  return reflection;
}


void KeyFrame::display() {
  cerr << "Displaying Frame" << endl;
  cerr << "toWaitTime=" << toWaitTime << endl;
  cerr << "waitTime=" << waitTime << endl;
  cerr << "toWaitTargets=" << toWaitTargets << endl;
  cerr << "maxWaitTime=" << maxWaitTime << endl;
  cerr << "Macros:\n";
  vector< shared_ptr<Macro> >::iterator it( macros.begin() ), end( macros.end() );
  for( ; it != end; ++it ) {
    (*it)->display();
  }
}

////////////////////////////////////////////////////////////////
// Skill
////////////////////////////////////////////////////////////////
Skill::Skill() {
  keyFrames.clear();
  reset();
  currentKeyFrame = -1; // set to invalid
}

void Skill::reset() {
  currentKeyFrame = 0;
  currentKeyFrameSet = false;
  currentKeyFrameSetTime = -1;
}

bool Skill::done(BodyModel *bodyModel, const double &currentTime){

  return ((currentKeyFrame == (int)(keyFrames.size() - 1)) && keyFrames[currentKeyFrame]->done(bodyModel, currentKeyFrameSetTime, currentTime));
}

void Skill::execute(BodyModel *bodyModel, const double &currentTime){

  // Check. Can be removed.
  if(done(bodyModel, currentTime)){
  // TODO: check why it prints it before kickoff - just a matter of initialization????
//    cerr << "Ideally, this should not be printed.\n";
    return ;
  }

  if(keyFrames[currentKeyFrame]->done(bodyModel, currentKeyFrameSetTime, currentTime)){
    currentKeyFrame += 1;
    currentKeyFrameSet = false;
    currentKeyFrameSetTime = -1;
  }

  if(!currentKeyFrameSet){
    keyFrames[currentKeyFrame]->execute(bodyModel);
    currentKeyFrameSet = true;
    currentKeyFrameSetTime = currentTime;
  }

  //cerr << "Skill::execute()  currentKeyFrame=" << currentKeyFrame << " currentTime=" << currentTime << "currentKeyFrameSetTime=" << currentKeyFrameSetTime << " done=" << keyFrames[currentKeyFrame]->done(bodyModel, currentKeyFrameSetTime, currentTime) << " currentKeyFrameSet=" << currentKeyFrameSet << endl;
}

void Skill::
appendKeyFrame( shared_ptr<KeyFrame> keyFrame ) {
  keyFrames.push_back( keyFrame );
  reset(); // just in case, we do it for each key frame added
}

shared_ptr<Skill> Skill::getReflection(BodyModel *bodyModel){

  shared_ptr<Skill> reflection(new Skill());

  reflection->currentKeyFrame = this->currentKeyFrame;
  reflection->currentKeyFrameSet = this->currentKeyFrameSet;
  reflection->currentKeyFrameSetTime = this->currentKeyFrameSetTime;

  (reflection->keyFrames).clear();
/*  for(int k = 0; k < this->keyFrames.size(); k++){
    reflection->appendKeyFrame(this->keyFrames[k]->getReflection(bodyModel));
  }
*/
  // TODO: find a better way to reflect - 
  // The problem with the commented code in the above line is:
  // When reflecting, it reflects left foot to right and right to left,
  // and thus the robot can start a skill in the right leg, instead of always
  // in the left. Below is a temp fix:
  int size = this->keyFrames.size();
  if (size % 2 != 0)
    throw "Uneven number of frames";
  for(int k = size / 2; k < size; k++){
    reflection->appendKeyFrame(this->keyFrames[k]->getReflection(bodyModel));
  }
  for(int k = 0; k < size / 2; k++){
    reflection->appendKeyFrame(this->keyFrames[k]->getReflection(bodyModel));
  }





  return reflection;
}

void Skill::display() {
  cerr << "Displaying skill" << endl;
  vector< shared_ptr<KeyFrame> >::iterator it( keyFrames.begin() ), end( keyFrames.end() );
  for( ; it != end; ++it ) {
    (*it)->display();
  }
}



////////////////////////////////////////////////////////////////
// TODO: remove this class kickSkill
////////////////////////////////////////////////////////////////
KickSkill::KickSkill( WorldModel* worldModel_, int& kickLeg_, VecPosition& ball_, VecPosition& kickDirection_, int& kickType_) :
  skillState( 0 ), kickTimeWait( -1 ),
  worldModel( worldModel_ ), kickLeg( kickLeg_ ),
  ball( ball_ ),
  kickDirection( kickDirection_ ),
  kickType( kickType_ ) {
}

void KickSkill::reset() {
  skillState = 0;
}

bool KickSkill::done(BodyModel *bodyModel, const double &currentTime){

  return (skillState == 50);
}

void KickSkill::execute(BodyModel *bodyModel, const double &currentTime){

  double KICK_WAIT_TIMEOUT = 3;

  if(skillState == 0){

    bodyModel->setInitialHead();
    bodyModel->setInitialArm(ARM_LEFT);
    bodyModel->setInitialArm(ARM_RIGHT);

    bodyModel->setInitialLeg(LEG_LEFT);
    bodyModel->setInitialLeg(LEG_RIGHT);
    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 1; 
    }
  }
  else if(skillState < 25){
    skillState++;
  }
  else if(skillState == 25){

    if(kickLeg == LEG_RIGHT){
      bodyModel->setTargetAngle(EFF_LL2, -10.0);
      bodyModel->setTargetAngle(EFF_LL6, 10.0);
    }
    else{
      bodyModel->setTargetAngle(EFF_RL2, 10.0);
      bodyModel->setTargetAngle(EFF_RL6, -10.0);
    }

    bodyModel->increaseTargetAngle(EFF_LL3, 10.0);
    bodyModel->increaseTargetAngle(EFF_LL4, -20.0);
    bodyModel->increaseTargetAngle(EFF_LL5, 10.0);

    bodyModel->increaseTargetAngle(EFF_RL3, 10.0);
    bodyModel->increaseTargetAngle(EFF_RL4, -20.0);
    bodyModel->increaseTargetAngle(EFF_RL5, 10.0);

    skillState = 26; 
  }
  else if(skillState == 26){
    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 27; 
    }
  }
  else if(skillState == 27){

    VecPosition temp = ball - (kickDirection * 0.18);
    temp.setZ(0.05);
    VecPosition localTarget = worldModel->g2l(temp);
    double angle = 0;

    if(kickLeg == LEG_RIGHT){

      if(!(bodyModel->reachOutLeg(LEG_RIGHT, localTarget, angle))){

        //localTarget = VecPosition(-0.02, -0.08, -0.29);
        //angle = 0;
        //bodyModel->reachOutLeg(LEG_RIGHT, localTarget, angle);

        //Equivalent
        bodyModel->setTargetAngle(EFF_RL1, 0);
        bodyModel->setTargetAngle(EFF_RL2, -10.0);
        bodyModel->setTargetAngle(EFF_RL3, 25.0);
        bodyModel->setTargetAngle(EFF_RL4, -95.0);
        bodyModel->setTargetAngle(EFF_RL5, 70.0);
        bodyModel->setTargetAngle(EFF_RL6, 10.0);

      }

      bodyModel->increaseTargetAngle(EFF_RL5, -10.0);
    }
    else{

      if(!(bodyModel->reachOutLeg(LEG_LEFT, localTarget, angle))){

        //	localTarget = VecPosition(-0.02, 0.08, -0.29);
        //	angle = 0;
        //	bodyModel->reachOutLeg(LEG_LEFT, localTarget, angle);

        //Equivalent
        bodyModel->setTargetAngle(EFF_LL1, 0);
        bodyModel->setTargetAngle(EFF_LL2, 10.0);
        bodyModel->setTargetAngle(EFF_LL3, 25.0);
        bodyModel->setTargetAngle(EFF_LL4, -95.0);
        bodyModel->setTargetAngle(EFF_LL5, 70.0);
        bodyModel->setTargetAngle(EFF_LL6, -10.0);

      }

      bodyModel->increaseTargetAngle(EFF_LL5, -10.0);
    }

    skillState = 28; 
  }
  else if(skillState == 28){
    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 29; 
    }
  }
  else if(skillState == 29){

    if(kickLeg == LEG_RIGHT){

      if(kickType == KICK_LONG){
        bodyModel->setScale(EFF_RL3, 2.0);
        bodyModel->setTargetAngle(EFF_RL3, 100.0);
      }
      else{//kickType == KICK_SHORT
        bodyModel->setTargetAngle(EFF_RL3, 60.0);
      }


      bodyModel->setTargetAngle(EFF_RL4, 0);
      bodyModel->setTargetAngle(EFF_RL5, 0);
    }    
    else{

      if(kickType == KICK_LONG){
        bodyModel->setScale(EFF_LL3, 2.0);
        bodyModel->setTargetAngle(EFF_LL3, 100.0);
      }
      else{//kickType == KICK_SHORT
        bodyModel->setTargetAngle(EFF_LL3, 60.0);
      }

      bodyModel->setTargetAngle(EFF_LL4, 0);
      bodyModel->setTargetAngle(EFF_LL5, 0);
    }

    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 30; 

      bodyModel->setScale(EFF_LL3, 1.0);
      bodyModel->setScale(EFF_RL3, 1.0);
    }
  }
  else if(skillState == 30){

    if(kickLeg == LEG_RIGHT){
      bodyModel->setTargetAngle(EFF_RL3, 40.0);
      bodyModel->setTargetAngle(EFF_RL4, -60.0);
      bodyModel->setTargetAngle(EFF_RL5, 20.0);
    }    
    else{
      bodyModel->setTargetAngle(EFF_LL3, 40.0);
      bodyModel->setTargetAngle(EFF_LL4, -60.0);
      bodyModel->setTargetAngle(EFF_LL5, 20.0);
    } 
    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 31; 
    }
  }
  else if(skillState == 31){

    bodyModel->setInitialHead();
    bodyModel->setInitialArm(ARM_LEFT);
    bodyModel->setInitialArm(ARM_RIGHT);

    bodyModel->setInitialLeg(LEG_LEFT);
    bodyModel->setInitialLeg(LEG_RIGHT);
    if (kickTimeWait < 0) {
      kickTimeWait = worldModel->getTime();
    }
    if(bodyModel->targetsReached() || 
        (kickTimeWait >= 0 && worldModel->getTime()-kickTimeWait > KICK_WAIT_TIMEOUT)){
      kickTimeWait = -1;
      skillState = 32; 
    }
  }
  else if(skillState < 50){
    skillState++;
  }
}

void KickSkill::display() {
  cerr << "Displaying KickSkill" << endl;
  cerr << "skillState=" << skillState << endl;
}
