#include<stdlib.h>
#include "Particle.h"

Particle::Particle() {

  M_prob = -1;
  resetObservations();
}

Particle::Particle( double prob,
		    Point2D pos,
		    AngRad ang,
		    Rectangle boundary ) {

  setParticle( prob, pos, ang, boundary );
  resetObservations();
}

double Particle::getProbability()
{
  return M_prob;
}

Point2D Particle::getPosition()
{
  return M_pos;
}

AngRad Particle::getAngle()
{
  return M_ang;
}


void Particle::setProbability( double prob )
{
  const double MIN_PROB = 1e-5;

  M_prob = ( prob < MIN_PROB ) ? MIN_PROB : ( prob > 1 ) ? 1 : prob;  
}



void Particle::degradeProbability( double factor )
{
  setProbability( M_prob * factor );
}

bool Particle::setPosition( Point2D pos, Rectangle boundary )
{
  if ( boundary.isInside( pos ) ) {
    M_pos = pos;
    return true;
  }
  return false;
}


void Particle::setAngle( AngRad ang )
{
  M_ang = normalizeAngle( ang );
}


void Particle::setParticle( double prob, Point2D pos, AngRad ang,
			    Rectangle boundary )
{
  setProbability( prob );
  if ( !setPosition( pos, boundary ) ){
    cout << "setParticle position error, tried to set " << pos << endl;
    M_pos = boundary.getPosOutside();
  }
  setAngle( ang );
}



void Particle::moveRelative( Vector2D relativeDisplacement,
			     AngRad rotation, Rectangle boundary )
{
  Point2D newPos =
    relativeDisplacement.relativeToGlobal( M_pos, M_ang );
  setPosition( newPos, boundary );
  setAngle( M_ang + rotation );
}

void Particle::moveGlobal( Vector2D globalDisplacement,
			   AngRad rotation, Rectangle boundary )
{
  setPosition( M_pos + globalDisplacement, boundary );
  setAngle( M_ang + rotation );
}


void Particle::randomWalk( double maxDistance, AngRad maxRotation,
			   Rectangle boundary )
{
  Vector2D dPos( maxDistance * ( 1 - M_prob ) * (2 * drand48() - 1),
		 maxDistance * ( 1 - M_prob ) * (2 * drand48() - 1) );
  AngRad dAng =  maxRotation * ( 1 - M_prob ) * (2 * drand48() - 1);
  
  moveGlobal( dPos, dAng, boundary );
}

void Particle::placeRandomly( Rectangle boundary )
{
  Vector2D dPos( boundary.getWidth() * drand48(),
		 boundary.getLength() * drand48() );
  setPosition( boundary.getBottomLeft() + dPos,
	       boundary );
  setAngle( TWOPI * drand48() - M_PI );
}



double Particle::getDistanceToPoint( Point2D p )
{
  return M_pos.getDistanceTo( p );
}

AngRad Particle::getBearingToPoint( Point2D p )
{
  return M_pos.getBearingTo(p, M_ang);
}


Point2D Particle::getNearestPointOnLine( Line2D l )
{
  // find point on line closest to particle position (M_pos)
  return l.getPointOnLineClosestTo( M_pos );
}



/*! Overloaded version of the C++ output operator for output. 
  \param os output stream to which information should be written
  \param p a Particle which must be printed
  \return output stream containing (x,y) */
ostream& operator << ( ostream &os, Particle p )
{
  return ( os << "( " << p.M_pos << ", " << Rad2Deg(p.M_ang) << ", " << p.M_prob << " )" );
}




void Particle::resetObservations(){
  for (int i = 0; i < NUM_WORLD_OBJS; i++) {
    M_expectedObs[i] = 0;
    M_seenObs[i] = 0;
  }
}

int Particle::getNumFramesExpected(int i){
  return M_expectedObs[i];
}

int Particle::getNumFramesNotSeen(int i){
  return M_seenObs[i];
}

void Particle::incrNumFramesExpected(int i){
  M_expectedObs[i]++;
}

void Particle::incrNumFramesNotSeen(int i){
  M_seenObs[i]++;
}

void Particle::incrNumFramesNotSeenForAll(){
  for (int i = 0; i < NUM_WORLD_OBJS; i++)
    M_seenObs[i]++;
}

void Particle::resetNumFramesExpected(int i){
  M_expectedObs[i] = 0;
}

void Particle::resetNumFramesNotSeen(int i){
  M_seenObs[i] = 0;
}
