#ifndef BODY_MODEL_H
#define BODY_MODEL_H

#include "headers.h"
#include "vecposition.h"
#include "hctmatrix.h"
#include "worldmodel.h"

#include <sstream>

struct Joint{
  
  double angle;

  Joint(){
    angle = 0;
  }
};

struct Effector{

  double minAngle;
  double maxAngle;
  
  double currentAngle;
  double targetAngle;
        
  // Constants for PID control
  double k1, k2, k3;
        
  // Error terms
  double currentError;// corr. k1
  double cumulativeError;// corr. k2
  double previousError;// corr. k3
        
  // set point tolerance
  double errorTolerance;

  double scale;

  Effector(const double &minAngle, const double &maxAngle, const double &k1, const double &k2, const double &k3, const double &errorTolerance){

    this->minAngle = minAngle;
    this->maxAngle = maxAngle;
  
    currentAngle = 0;
    targetAngle = 0;
        
    this->k1 = k1;
    this->k2 = k2;
    this->k3 = k3;
     
    resetErrors();
    this->errorTolerance = errorTolerance;

    scale = 1.0;

  }

  void resetErrors(){
    currentError = 0;
    previousError = 0;
    cumulativeError = 0;
  }
        
  void updateErrors(){

    previousError = currentError;
    currentError = targetAngle - currentAngle;
    cumulativeError += currentError;

  }

  void setTargetAngle(const double &angle){
    
    if(angle < minAngle){
      targetAngle = minAngle;
    }
    else if(angle > maxAngle){
      targetAngle = maxAngle;
    }    
    else{
      targetAngle = angle;
    }
    
  }

  void update(const double &angle){

    currentAngle = angle;
  }
        
  bool targetReached(){
    
    return (fabs(targetAngle - currentAngle) < errorTolerance);
  }
};

struct Component{
  
  int parent;
  
  double mass;
  VecPosition translation;
  VecPosition anchor;
  VecPosition axis;
  
  HCTMatrix transformFromParent;
  HCTMatrix transformFromRoot;
  
  HCTMatrix translateMatrix;
  HCTMatrix backTranslateMatrix;

  Component(const int &parent, const double &mass, const VecPosition &translation, const VecPosition &anchor, const VecPosition &axis){
    
    this->parent = parent;
    this->mass = mass;
    
    this->translation = translation;
    this->anchor = anchor;
    this->axis = axis;
    
    transformFromParent = HCTMatrix();
    transformFromRoot = HCTMatrix();
    
    translateMatrix = HCTMatrix(HCT_TRANSLATE, VecPosition(translation) + VecPosition(anchor));
    backTranslateMatrix = HCTMatrix(HCT_TRANSLATE, -VecPosition(anchor));
  }
};


class BodyModel{
  
 private:
  
  WorldModel *worldModel;
  
  // Joints
  std::vector<Joint> joint;
  
  // Effectors
  std::vector<Effector> effector;
  
  // Components
  std::vector<Component> component;
  
  VecPosition gyroRates;
  VecPosition accelRates;
  vector<VecPosition> FRPCentre;
  vector<VecPosition> FRPForce;
  
  HCTMatrix bodyWorldInterface;
  
  HCTMatrix ll1, ll2, ll3, ll4, ll5;
  HCTMatrix rl1, rl2, rl3, rl4, rl5;

  double fallAngle;

  void refreshTorso();
  void refreshComponent(const int &index);
  void refreshHead();
  void refreshLeftArm();
  void refreshRightArm();
  void refreshLeftLeg();
  void refreshRightLeg();
  
  inline void computeFallAngle(){

    VecPosition zAxis = VecPosition(0, 0, 1.0);
    VecPosition bodyZAxis = (worldModel->l2g(VecPosition(0, 0, 1.0)) - worldModel->l2g(VecPosition(0, 0, 0))).normalize();
    fallAngle = VecPosition(0, 0, 0).getAngleBetweenPoints(zAxis, bodyZAxis);
  }

 public:
  
  BodyModel(WorldModel *worldModel);
  ~BodyModel();
  
  void initialiseEffectors();
  void initialiseComponents();
  
  inline double getJointAngle(const int &i){ return joint[i].angle; }
  inline void setJointAngle(const int &i, const double &a){ joint[i].angle = a; }
  
  void display();
  void displayDerived();
  
  inline double getCurrentAngle(const int &EffectorID){ return effector[EffectorID].currentAngle; }
  inline void setCurrentAngle(const int &EffectorID, const double &angle){ effector[EffectorID].currentAngle = angle; }

  inline double getTargetAngle(const int &EffectorID){ return effector[EffectorID].targetAngle; }
  inline void setTargetAngle(const int &EffectorID, const double &angle){ effector[EffectorID].setTargetAngle(angle); }
  inline void increaseTargetAngle(const int &EffectorID, const double &increase){ effector[EffectorID].setTargetAngle(effector[EffectorID].targetAngle + increase); };
  
  inline double getScale(const int &EffectorID){ return effector[EffectorID].scale; }
  inline void setScale(const int &EffectorID, double s){ effector[EffectorID].scale = s; }
  
  inline bool getTargetReached(const int &EffectorID){ return effector[EffectorID].targetReached(); }
  
  double computeTorque(const int &effectorID);
  
  inline VecPosition getGyroRates(){ return gyroRates; } 
  inline void setGyroRates(const double &rateX, const double &rateY, const double &rateZ){ gyroRates = VecPosition(rateX, rateY, rateZ); }

  inline VecPosition getAccelRates(){ return accelRates; } 
  inline void setAccelRates(const double &rateX, const double &rateY, const double &rateZ){ accelRates = VecPosition(rateX, rateY, rateZ); }
  
  inline VecPosition getFRPCentre(const int &legIndex){ return FRPCentre[legIndex]; }
  inline VecPosition getFRPForce(const int &legIndex){ return FRPForce[legIndex]; }
  inline void setFRP(const int &legIndex, const VecPosition &centre, const VecPosition &force){
    
    FRPCentre[legIndex] = VecPosition(centre);
    FRPForce[legIndex] = VecPosition(force);
  }
  
  void refresh();

  bool reachOutLeg(const int &legIndex, const VecPosition &localTarget, const double &angle);
  bool targetsReached();

  VecPosition getFootCG(const int &legIndex, const double &ang1, const double &ang2, const double &ang3, const double &ang4, const double &ang5, const double &ang6);

  VecPosition getFootCG(const int &legIndex);

  void getFootTransform(const int &legIndex, HCTMatrix &m, const double &ang1, const double &ang2, const double &ang3, const double &ang4, const double &ang5, const double &ang6);

  void getFootGlobalCGAndAxes(const int &legIndex, VecPosition &CG, VecPosition &xAxis, VecPosition &yAxis, VecPosition &zAxis);

  bool legInverseKinematics(const int &legIndex, const VecPosition &localTarget, const double &footAngle, double &a1, double &a1, double &a3, double &a4, double &a5, double &a6);

  inline double getFallAngle(){ return fallAngle; }

};

#endif // BODY_MODEL_H


