#!/usr/bin/env ruby

# Cross-entropy parameter generator for 
# RoboCup 3D-Sim learner
# 
# This script takes in an experiment name and the current iteration (beginning at 0)
# and generates the next population
#
# Yinon Bentor
require 'pp'

$debug = true

$allparams = {
  "agent2beamX"         => { :init_mu => -0.718, :init_sigma => 0.001, :max => -0.55, :min => -0.85, },
  "agent2beamY"         => { :init_mu => -0.125, :init_sigma => 0.001, :max => 0.5, :min => -0.30, },
  "agent2beamAngle"     => { :init_mu => 5.2, :init_sigma => 1.0, },
  "returnKickScale"     => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL1" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL2" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL3" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL4" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL5" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL6" => { :init_mu => 1.0, :init_sigma => 1.0, :min => 0.250, :max => 5.0},
  "targetHeight"        => { :init_mu => 0.05, :init_sigma => 0.005, :max => 0.065, :min => 0.040},
  "angleAtKick"         => { :init_mu => 0.0, :init_sigma => 5.0, },
}

# Experiment information should be given on command line:
if(ARGV.size != 3) 
  STDERR.print "Usage: #{$0} <experiment_path> <current_iter> <pop_size>\n"
  exit(1)
end

$experiment_path = ARGV.shift
$current_iter = ARGV.shift.to_i
$pop_size = ARGV.shift.to_i

## Global params
$count_top = ($pop_size * 0.10).to_i;  # number of individuals to base new parameter on

#initialize the parameters with mu and sigma values:

# Initiate params for first generation to draw from same distribution 
$initial_params = {}
$allparams.keys.each do |i|
  $initial_params[i] = {
    :mu => $allparams[i][:init_mu],
    :sigma => $allparams[i][:init_sigma]
  }
end

# Define some methods that generate random numbers from different
# PDFs
module Stats
  def Stats.sum(nums)
    nums.inject(0) { |a,x| x + a } 
  end
  
  def Stats.mean(nums)
     sum(nums).to_f/nums.size 
  end
  
  def Stats.variance(nums)
    squares = sum(nums.map{|i| i*i}).to_f
    squares/nums.size - mean(nums)**2
  end
  
  
  # This method generates a sample from a normal distribution with 
  # mean 0.0 and variance 1.0 using a standard numerical recipe.  
  #
  # This function uses the Box-Muller transform for generating Gaussian
  # pseudo numbers from a uniform pseudo number generator, per:
  # http://www.taygeta.com/random/gaussian.html
  #
  # It is copied verbatim from: http://snippets.dzone.com/posts/show/4697
  def Stats.normal_rand 
     u1 = u2 = w = g1 = g2 = 0
     
     begin
       u1 = 2 * rand - 1
       u2 = 2 * rand - 1
       w = u1 * u1 + u2 * u2
     end while w >= 1
     
     w = Math::sqrt( ( -2 * Math::log(w)) / w )
     g2 = u1 * w;
     g1 = u2 * w;
     # g1 is returned  
  end
  
  # Return a single sample from the Gaussian distribution
  # with mean _mu_ and variance _sigma_
  def Stats.gaussian_rand(mu, sigma) 
    sigma * normal_rand() + mu
  end
end

module Generator
  def Generator.read_value(generation, instance)
    valuedata = File.open($experiment_path + "/value_#{generation}_i_#{instance}.txt").read
    valuedata.to_f
  end
  
  # Answers a hash of the parameters listed in the plaintext file 
  # /params_#{generation}_i_#{instance}.txt, which should be separated by newlines 
  # and have the form #{key}\t#{value}
  #
  # This is the format written by #write_params()
  def Generator.read_params(generation, instance)
    result = {}
    paramdata = File.open($experiment_path + "/params_#{generation}_i_#{instance}.txt").read
    paramdata.split(/\n/).each do |line|
      (key, value) = line.split(/\t/)
      if(value)
        result[key.to_s] = value.to_f
      end
    end
    print "Read params from #{$experiment_path + "/params_#{generation}_i_#{instance}.txt"}\n"
    return result
  end
  
  def Generator.write_params(params, generation, instance) 
#    if $debug then STDERR.print "writing to: " + $experiment_path + "/params_#{generation}_i_#{instance}.txt" 

    outfile = File.open($experiment_path + "/params_#{generation}_i_#{instance}.txt", 'w');
    params.sort_by {|k| k.first}.each do |pair|
       k = pair.first
       v = pair.last
#      if $debug then STDERR.print "  #{k}\t#{v}\n"
      outfile.write("#{k}\t#{v}\n");
    end
    outfile.close
  end
  
  # Generate a population with _count_ elements composed of the given
  # parameters.  _params_ should be given as a dictionary from 
  # parameter name to a hash with keys :mu and :sigma
  #
  # Returns a list of individuals as dictionarys of parameters names
  # (strings) to real values.
  def Generator.generate_pop(count, params)
    result = []
    count.times do
      individual = {}
      params.each_pair do |k, v| 
        val = Stats::gaussian_rand(v[:mu],v[:sigma] + Generator::noise()*$allparams[k][:init_sigma])
	if ($allparams[k][:max] != nil && val > $allparams[k][:max]) then
	  val = $allparams[k][:max]
	end
	if ($allparams[k][:min] != nil && val < $allparams[k][:min]) then
	  val = $allparams[k][:min]
 	end
	individual[k] = val
      end
      result << individual
    end
    result
  end
  
  # Answer a double representing the annealed noise in sigma^2 
  # corresponding to the variable Z in Szita and Lorincz's paper
  def Generator.noise() 
    noise = 1.0 - $current_iter.to_f/2.0
    
    return [noise, 0].max
  end
  
  # Using the old set of parameters and associated values
  # from the objective function, return new set of individuals
  # for the next generation
  # 
  # The valued_params array of hashes should look like this:
  #  [{ :params => {hash of parameters}
  #     :value  => the result of the objective function, should be orderable
  #  }]
  #
  # Parameter _n_ is the number of elements to use to base
  # the new population on
  def Generator.new_params(valued_params, n = 10)
    fmt = "%0.3f"
    
    base = valued_params.sort_by{|i| i[:value]}.first(n)
    print "Best value: #{fmt % base.first[:value]}\tparams: #{base.first[:params].inspect}\n"
    newparams = {}
    base.first[:params].keys.sort.each do |k|
      # _values_ is an array of real values of the parameter for the top
      # _n_ individuals
      x = base.map{|i| i[:params][k]}
      print " #{k}:\tVariance: #{fmt % Stats::variance(x)}\tMean: #{fmt % Stats::mean(x)}\n"
      newparams[k] = {:mu => Stats::mean(x), :sigma => Stats::variance(x) ** 0.5} 
    end
    Generator::generate_pop(valued_params.size, newparams)
  end
end


#write out the population parameters to a set of files
def write_pop(population)
  population.each_with_index do |e, index|
    Generator.write_params(e, $current_iter+1, index)
  end
end

if($current_iter == 0)
  #initialize pop
  start = Generator::generate_pop($pop_size, $initial_params)
  write_pop(start)
else
  #read in pop and values from files
  result = []
  Dir[$experiment_path + "/value_#{$current_iter}_i_*.txt"].each do |file|
    file =~ /value_#{$current_iter}_i_([0-9]+).txt$/
    ind = $1.to_i
    result << { :params => Generator.read_params($current_iter, ind),
                :value => Generator.read_value($current_iter, ind) }
  end
  nextpop = Generator::new_params(result, $count_top)
  write_pop(nextpop)
end
