#!/usr/bin/env ruby

# Policy-gradient parameter generator for RoboCup 3-D Sim learner
#
# Generates a new set of parameters for policy gradient learning.
# If this is the first iteration, generates an initial policy and
# random perturbations of it.  If this is a subsequent iteration, 
# evaluate the previous iteration to generate a next policy and 
# then generate random perturbations of that policy.  We are following
# the algorithm of Nate Kohl and Peter Stone, in 
# "Machine Laerning for Fast Quadrupedal Locomotion" (AAAI 2004).
#
# Usage: generator.rb <experiment_path> <current_iter> <pop_size>
#
# Yinon Bentor
# 2009-12-07

# Experiment information should be given on command line:
require 'pp'
require 'yaml'

if(ARGV.size != 3) 
  STDERR.print "Usage: #{$0} <experiment_path> <current_iter> <pop_size>\n"
  exit(1)
end

$experiment_path = ARGV.shift
$current_iter = ARGV.shift.to_i
$pop_size = ARGV.shift.to_i
# we use this value for the fitness function if no policies were evaluated on a particular dimension
VERY_NEGATIVE = -1000  
$eta = 0.05
$annealing_rate = 0.8
$epsilon = 0.1 # 0.1m 
$actions = [:add, :sub, :zero]


$allparams = {
  # each parameter must have a :start value, which is the value in the initial policy and
  # a :e_j value, which is the amount by which the parameter can be randomly perturbated at 
  # each iteration.  Optionally, a :max and :min value can specify limits on the parameter.
  "agent2beamX"         => { :start => -0.718, :e_j => 0.005, :max => -0.55, :min => -0.85, },
  "agent2beamY"         => { :start => -0.125, :e_j => 0.005, :max => -0.09, :min => -0.14, },
  "agent2beamAngle"     => { :start => 5.2,    :e_j => 0.05, },
  "returnKickScale"     => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL1" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL2" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL3" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL4" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL5" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "forwardKickScaleLL6" => { :start => 1.0,    :e_j => 0.1, :min => 0.250, :max => 5.0},
  "targetHeight"        => { :start => 0.05,   :e_j => 0.005, :max => 0.065, :min => 0.040},
  "angleAtKick"         => { :start => 0.0,    :e_j => 0.05, },
}

# Persist the policy to a file.  The policy should be a 
# hash 
def write_policy(filename, policy)
  outfile = File.open($experiment_path + "/" + filename, 'w');
  policy.sort_by {|k| k.first}.each do |pair|
    k = pair.first
    v = pair.last
    outfile.write("#{k}\t#{v}\n");
  end
  outfile.close
end

def read_policy(filename)
  result = {}
  paramdata = File.open($experiment_path + "/" + filename).read
  paramdata.split(/\n/).each do |line|
    (key, value) = line.split(/\t/)
    if(value)
      result[key.to_s] = value.to_f
    end
  end
  print "Read policy from #{$experiment_path}/#{filename}\n"
  return result
end

# reads all the available scores for a given iteration and returns them as
# a map from policy # to score.  Higher scores are better.  It is possible that
# not all policies returned scores, in which case there will be no entry in
# the returned map for that policy #
def read_scores(iter)
  result = {}
  Dir[$experiment_path + "/value_#{iter}_i_*.txt"].each do |file|
    file =~ /value_#{$current_iter}_i_([0-9]+).txt$/
    ind = $1.to_i
    result[ind] = File.open(file).read.to_f
  end
  return result
end

last_policy = {}

# write out the initial policy as the "last" policy if 
# this is the first iteration
if ($current_iter == 0)
  finalpolicy = {}
  $allparams.each_pair do |k, v|
    finalpolicy[k] = v[:start]
  end
  
  next_policy = finalpolicy
  print "next policy:\n"
  pp next_policy
  
  #write out $allparams as the initial parameters:
  File.open("#{$experiment_path}/init_0.yml", 'w') {|f| f.write($allparams.to_yaml)}
  
  write_policy("finalpolicy_0.txt", finalpolicy)
else
  # compute the final policy for the last iteration, now that we've evaluated:
  
  scores = read_scores($current_iter)
  pp scores
  
  #$allparams = YAML::load_file("#{$experiment_path}/init_#{$current_iter}.yml")
  last_clusters = YAML::load_file("#{$experiment_path}/clusters_#{$current_iter}.yml")
  pp last_clusters
  
  a = {}

  # compute avg_{+e}, avg_0, avg_{-e} for each parameter:
  $allparams.keys.each do |param|
        
    #initialize a's
    a[param] = {}
    $actions.each do |action|
      a[param][action] = {
        :sum => 0.0,
        :count => 0
      }
    end
    
    scores.each_pair do |i, val|
      $actions.each do |action|
        if (last_clusters[i][param] == action)
          a[param][action][:count] = a[param][action][:count] + 1
          a[param][action][:sum] = a[param][action][:sum] + val
        end
      end
    end

    #now average them
    $actions.each do |action|
      if (a[param][action][:count] == 0)
        a[param][action][:avg] = VERY_NEGATIVE
      else
        a[param][action][:avg] = a[param][action][:sum] / (a[param][action][:count].to_f)
      end
    end    
    
  end
  print "A: "
  pp a 
  
  #compute the final policy from this iteration:
  
  delta = {} # the A vector in the policy gradient algorithm
  $allparams.keys.each do |param|
    if ((a[param][:zero][:avg] > (a[param][:add][:avg] - $epsilon)) &&
        (a[param][:zero][:avg] > (a[param][:sub][:avg] - $epsilon)))
      delta[param] = 0
      # we didn't modify the parameter in this direction, so we should anneal it
      #print "Annealing parameter #{param} from #{$allparams[param][:e_j]} to "
      #$allparams[param][:e_j] = $allparams[param][:e_j] * $annealing_rate 
      #print "#{$allparams[param][:e_j]}\n"

    else
      delta[param] = a[param][:add][:avg] - a[param][:sub][:avg]      
    end
  end
  print "DELTA: "
  pp delta
  
  
  # compute the magnitude of the delta vector:
  mag = 0.0
  $allparams.keys.each do |param|
    mag = mag + (delta[param]**2.0)
  end
  print "mag, pre-sqrt: #{mag}\n"
  mag = Math.sqrt(mag)

  last_policy = read_policy("finalpolicy_#{$current_iter-1}.txt")
  next_policy = {}
  
  if (mag == 0.0)
    print "Policy didn't change\n"
    next_policy = last_policy
  else 
    $allparams.keys.each do |param|
      next_policy[param] = last_policy[param] + (delta[param]/mag) * $eta
      
      #observe mins and maxes:
      if ($allparams[param][:max] != nil && next_policy[param] > $allparams[param][:max])
        print "Was going to set #{param} to #{next_policy[param]}, but bounded to max #{$allparams[param][:max]}\n"
        next_policy[param] = $allparams[param][:max]
      elsif ($allparams[param][:min] != nil && next_policy[param] < $allparams[param][:min])
        print "Was going to set #{param} to #{next_policy[param]}, but bounded to min #{$allparams[param][:min]}\n"
        next_policy[param] = $allparams[param][:min]
      end
    end
  end
  print "Next policy: "
  pp next_policy
  write_policy("finalpolicy_#{$current_iter}.txt", next_policy)
end

# compute the next set of random policies

clusters = {}

(0...$pop_size).each do |i|
  clusters[i] = {}
  policy = {}
  $allparams.each_pair do |k, v|
    # figure out if we're going to keep this param the same in this random policy,
    # subtract e_j from it, or add e_j to it
    action = $actions[rand*3]
    if (action == :add)
      delta = $allparams[k][:e_j]
    elsif (action == :sub) 
      delta = -1.0 * $allparams[k][:e_j]
    else
      delta = 0.0
    end
    
    policy[k] = next_policy[k] + delta
    clusters[i][k] = action
    write_policy("params_#{$current_iter+1}_i_#{i}.txt", policy)
  end
end

# write out clusters to YAML file:
File.open($experiment_path + "/clusters_#{$current_iter+1}.yml", 'w') {|f| f.write(clusters.to_yaml)}

# and the (possibly annealed) $allparams
File.open("#{$experiment_path}/init_#{$current_iter+1}.yml", 'w') {|f| f.write($allparams.to_yaml)}
